/*******************************************************************************

    TERMMETRO C/ DISPLAYS DE 7 SEG (4 DGITOS)

    Para a faixa de: 0 - 1023,75 C (TERMOPAR TIPO 'K')

    Com PIC 16f628A e MAX 6675

    Autor: Claudio Larios

    Data: 26/06/2020

    Com 4 Displays de 7 seguimentos (opcionalmente ctodo ou nodo comum)
    ImportAnte: Use o esquema correto conforme a opo acima.

    Obs. Montagem de natureza didtica

    Este arquivo  parte integrante do blog Picsource

*******************************************************************************/

//    #define ANODO_COMUM  

// Comente o acima para catodo comum e descomente para nodo comum
//=============================================================================/


#include <stdio.h>
#include <stdlib.h>
#include <string.h> //para usar funoes de string deve se adicionar este header
#include <xc.h>
#define _XTAL_FREQ 4000000//frequncia usada (osc. interno))

// PALAVRA DE CONFIGURAO DO PIC
#pragma config FOSC = INTOSCIO      // Oscillator Selection bits 
#pragma config WDTE = OFF       // Watchdog Timer Enable bit (WDT disabled)
#pragma config PWRTE = OFF      // Power-up Timer Enable bit (PWRT disabled)
#pragma config MCLRE = ON       // RA5/MCLR/VPP Pin Function Select bit (RA5/MCLR/VPP pin function is MCLR)
#pragma config BOREN = ON       // Brown-out Detect Enable bit (BOD enabled)
#pragma config LVP = OFF        // Low-Voltage Programming Enable bit (RB4/PGM pin has digital I/O function, HV on MCLR must be used for programming)
#pragma config CPD = OFF        // Data EE Memory Code Protection bit (Data memory code protection off)
#pragma config CP = OFF         // Flash Program Memory Code Protection bit (Code protection off)


//define pinos de comunicao com o Max6675
#define   so       RA7
#define   sck      RA6
#define   cs       RA4
//constantes para converso 7 seguimentos

//                         PGFEDCBA   SEGUIMENTOS
//                   PORTB 76543210
const char NUM[] = {     0B00111111,//0
                         0B00000110,//1
                         0B01011011,//2
                         0B01001111,//3
                         0B01100110,//4
                         0B01101101,//5
                         0B01111101,//6
                         0B00000111,//7
                         0B01111111,//8
                         0B01101111, //9
                         0B01010100, //'n' 10
                         0B01011100, //'o' 11
                         0B01111001, //'E' 12
                                        };

 //variveis globais
 char dig3=0,dig2=0,dig1=0,dig0=0;
 char dig3a=0,dig2a=0,dig1a=0,dig0a=0,digdec=0,digcent=0;
 char dlyread=0;
 char dlyread1=0;
 char dlympx=0;
 char vez=0;
 int  valor=0;
 bit fok=0 ;
 bit f_conv_ok=0; //flag de converso no realizada
 bit fsts=0;//flag status 'existe sensor ou no ligado'
 bit fp1=0, fp2=0;

 /******************************************************************************
           MULTIPLEXA OS DISPLAYS
******************************************************************************/

 void multiplexa(){    
    
    dlympx++;
    if(dlympx<125)return;//multiplexa 1 em cada 125 vezes
    dlympx=0;

    if(fok){
     dig0=dig0a;
     dig1=dig1a;
     dig2=dig2a;
     dig3=dig3a;
     fok=0;
    }

   if(++vez>3) vez=0;
    PORTA&=0xf0;//apaga display (resetando)
    #ifdef ANODO_COMUM
    PORTB=0xff;
    #else
    PORTB=0;
    #endif

    switch (vez){

     case 0: {
       PORTB=NUM[dig3];
       #ifdef ANODO_COMUM
       PORTB=~PORTB;
       #endif
       if(dig3) PORTA= PORTA|8;//apaga zeros a esquerda
       break;}
     case 1: {
       PORTB=NUM[dig2];
       if(fp2)PORTB=PORTB|0X80;
       #ifdef ANODO_COMUM
       PORTB=~PORTB;
       #endif
       if(dig3 || dig2) PORTA= PORTA|4;//apaga zeros a esquerda       
       break;}
     case 2: {
       PORTB=NUM[dig1];
       if(fp1)PORTB=PORTB|0X80;
       #ifdef ANODO_COMUM
       PORTB=~PORTB;
       #endif
       if(dig3 || dig2 || dig1) PORTA= PORTA|2;//apaga zeros a esquerda
       break;}
     case 3: {
       PORTB=NUM[dig0];
       #ifdef ANODO_COMUM
       PORTB=~PORTB;
       #endif
       PORTA= PORTA|1; break;}
    }
 }

//==============================================================================
//         ROTINA DE LEITURA DOS DADOS DO MAX6675
//==============================================================================

unsigned int  le_max6675(){
  char a;
  int  dado;   
  #define DELAY_ENTRE_CONVERCOES 50 //Aprox. 0,5 seg
   //temporiza de forma dinmica os intervalos necessrios para a converso
   //mnimo 250 ms
     dlyread++;
     if(dlyread==0){
        dlyread1++;
        if(dlyread1<DELAY_ENTRE_CONVERCOES){         
          return 0xffff;
        }
     }
     else{ return 0xffff;}
    dlyread1=0;
   //inicia a comunicao com o Max6675
    dado=0;
    cs=0;
    for(a=0;a<16;a++){//16 bits
      sck=1;//clock=1
      dado=dado<<1;//desloca para esquerda
      if(so)dado=dado|1;//se '1' seta bit0 de dado
      sck=0;//clock=0
    }

   cs=1;//libera para novas converes de temperatura


   if(dado & 4)fsts=1; else fsts=0;//termopar status: 1 = aberto e 0 = conectado
   dado=dado>>3;//ajusta para direita 3 casa

   f_conv_ok=1;//sucesso na convero da temperatura
   return dado;

}
//==============================================================================
//     ROTINA DE APRESENTAO DOS RESULTADOS NOS DISPLAYS DE 7 SEGUIMENTOS
//==============================================================================
void mostra_nos_displays(int  value){
    if(f_conv_ok==1){//se '1' ento  tem uma convero
       f_conv_ok=0;
       digcent=0;
       digdec=0;

       if(valor & 1)digcent=1;//obtem valor da segunda casa ps vrgula
       if(valor & 2)digdec=1;//obtm valor da primeira casa ps vrgula

       if((digcent==1) && (digdec==1)){//converte p/ valores corretos
         digdec=7;
         digcent=5;
       }
       else if((digcent==1) && (digdec==0)){//converte p/ valores corretos
         digdec=5;
         digcent=0;
       }
       else if((digcent==0) && (digdec==1)){//converte p/ valores corretos
         digdec=2;
         digcent=5;
       }

       valor=valor>>2;//divide por 4 (lsb=0,25)

       if(valor>=1000){ //valores iguais ou acima de 1000 graus C     
         dig3a=valor/1000;//converte para decimal nos dgitos auxiliares
         valor=valor%1000;     
         dig2a=valor/100;
         valor=valor%100;      
         dig1a=valor/10;
         dig0a=valor%10;
         fp2=0;//desliga ponto do dgito 2
         fp1=0;//desliga ponto do dgito 1        
       }
       else{
         if(valor>=100){//valores iguais ou acima de 100 graus C   
           dig3a=valor/100;
           valor=valor%100;      
           dig2a=valor/10;
           dig1a=valor%10;
           dig0a=digdec;
           fp1=1;//liga ponto do dgito 1
           fp2=0;             
         }
         else{//valores abaixo de 100 graus C   
             dig3a=valor/10;
             dig2a=valor%10;
             dig1a=digdec;
             dig0a=digcent;             
             fp1=0;//desliga ponto
             fp2=1; //liga ponto do dgito 2         
         }      
      }      
      
       if(fsts==1){//verifica se o termopar est interrompido ou desconectado
       dig3a=10;//'n' envia mensagem 'noSE' 
       dig2a=11;//'o'
       dig1a=5; //'S'
       dig0a=12;//'E'
       fp1=0;//desliga ponto
       fp2=0; //desliga ponto do dgito 2   
     }

      fok=1;//avisa multiplexagem que tem uma leitura nova
     }
}

//==============================================================================
//         ROTINA MAIN
//==============================================================================

void main() {
 
   CMCON=7;//desliga comparadores internos
   TRISA=0b10100000;//sentido fixo das PORTAs
   TRISB=0;//idem
   PORTA&=0xf0;//desliga displays anodo ou catodo
   cs=1;
   sck=0; 

//==============================================================================
//         LAO DE REPETIO PRINCIPAL
//==============================================================================

   for(;;){
       
     valor=le_max6675();//le sensor
     mostra_nos_displays(valor); //converte valores e apresenta nos displays    
     multiplexa();//multiplexa os dgitos      
   }
}
//==============================================================================
//                         FIM DO PROGRAMA
//==============================================================================
